package TMApplet;

import java.applet.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;

/*
 * Controls the machine.
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@unsw.edu.au">peteg@unsw.edu.au</A>)
 * @version $Id: ControlPanel.java,v 1.1.1.1 2002/02/28 12:09:45 peteg Exp $
 */
class ControlPanel extends Panel {
    private TMApplet parent;

    // GUI components
    private Button btnReset = new Button("Reset"),
	btnRun = new Button("Run"),
	btnStop = new Button("Stop");
    private Choice choiceSpeed;

    public Insets getInsets() {
	return new Insets(5,5,5,5);
    }

    ControlPanel(TMApplet tmApplet, int defSpeed, Color buttonBG) {
	parent = tmApplet;

	// GridLayout, 1 row, 5 pixels around all edges of each cell.
	setLayout(new GridLayout(1,0,5,5));

	btnReset.addActionListener(new ActionListener() {
	    public void actionPerformed(ActionEvent e) {
		if(parent.thread != null)
		    return;

		for(int i = 0; i < parent.tapeHeads.length; i++)
		    parent.tapeHeads[i].reset();
		parent.states.reset();
		parent.states.update(parent.states.getGraphics());
		btnRun.setEnabled(true);
	    }
	});
 	btnReset.setBackground(buttonBG);
	this.add(btnReset);

	btnStop.addActionListener(new ActionListener() {
	    public void actionPerformed(ActionEvent e) {
		parent.thread = null;
		return;
	    }
	});
	btnStop.setBackground(buttonBG);
	btnStop.setEnabled(false);
	this.add(btnStop);

	btnRun.addActionListener(new ActionListener() {
	    public void actionPerformed(ActionEvent e) {
		if(parent.thread == null) {
		    parent.thread = new Thread(parent);
		    parent.thread.start();
		}
		return;
	    }
	});
	btnRun.setBackground(buttonBG);
	add(btnRun);
	
        choiceSpeed = new Choice();
	choiceSpeed.setBackground(buttonBG);
        choiceSpeed.addItem("Step");
        choiceSpeed.addItem("Slow");
        choiceSpeed.addItem("Medium");
        choiceSpeed.addItem("Fast");
        choiceSpeed.addItem("Crank");
	choiceSpeed.select(defSpeed);
        add(choiceSpeed);
    }

    public void activeButtons(boolean running, boolean halted) {
        if(running) {
	    btnRun.setEnabled(false);
	    btnReset.setEnabled(false);
            if(speed() != 0)
                btnStop.setEnabled(true);
        } else {
	    btnStop.setEnabled(false);
	    btnReset.setEnabled(true);
            if(!halted)
                btnRun.setEnabled(true);
            else
                btnRun.setEnabled(false);
        }
    }

    public int speed() {
        return(choiceSpeed.getSelectedIndex());
    }

    public void paint(Graphics g) {
	update(g);
    }

    public void update(Graphics g) {
	parent.lowered3DBorder(this.getGraphics(), getSize());
    }
}
