package TMApplet;

import java.applet.*;
import java.awt.*;
import java.util.*;

/*
 * TMApplet - a Turing machine simulator.<br>
 * Copyright (C) 1998  pete gammie <> peteg@unsw.edu.au
 * <p>
 * Comments: This applet is always incomplete, so you may want to check this
 * again sometime in the future.
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@unsw.edu.au">peteg@unsw.edu.au</A>)
 * @version $Id: TMApplet.java,v 1.1.1.1 2002/02/28 12:09:45 peteg Exp $
 */
public class TMApplet extends Applet implements Runnable {
    Thread thread;

    // Parameter names
    private static final String PAlphabet = "Alphabet";
    private static final String PBreakPoint = "BreakPoint";
    private static final String PHeadOffset = "HeadOffset";
    private static final String PNumTapes = "NumTapes";
    private static final String PState = "State";
    private static final String PTapeOffset = "TapeOffset";
    private static final String PTapeString = "TapeString";
    private static final String PThreeD = "ThreeD";

    private static final String PAutoStart = "AutoStart";
    private static final String PBGColour = "BGColour";
    private static final String PButtonBG = "ButtonBG";
    private static final String PDefSpeed = "DefSpeed";
    private static final String PDebug = "Debug";

    // GUI components
    ControlPanel controlPanel;
    States states;
    TapeHead[] tapeHeads;

    private Color buttonBG;
    Color bg;
    private boolean threeD;

    // Turing Machine variables
    private String alphabet = "";
    static final int Blank = 0;
    private boolean autoStart;
    private int defSpeed;
    private boolean debug = false;

    public Insets getInsets() {
	return new Insets(5,5,5,5);
    }

    public String getAppletInfo() {
	return "Turing Machine Applet <-> pete gammie, 1998.";
    }

    public String[][] getParameterInfo() {
	String info[][] = {
	    {PAlphabet,		"String",	"Alphabet used to encode tape(s) [first element is 'blank', '_' is the wildcard - default: alphanumeric] - If present, an initial \'\\\' is ignored."},
	    {PNumTapes,		"int",		"Number of tapes [default: 1]. The 'n' in the following parameters indicate which tape the initialisation applies to."},
	    {PHeadOffset + "n",	"int",		"Initial position of tape head [default: 0]."},
	    {PTapeOffset + "n",	"int",		"Offset where TapeString begins [default: 0]."},
	    {PTapeString + "n",	"String",	"Initial tape string [default: blank]."},
	    {PState + "1",	"String",	"State transition table entries, each containing [5|7] comma-delimited values:"},
	    {PState + "2",	"String",	"1. Current state name,"},
	    {PState + "3",	"String",	"2. Tape the following symbol is matched to [only for multiple tapes],"},
	    {PState + "4",	"String",	"3. Current symbol on the tape,"},
	    {PState + "5",	"String",	"4. Tape the following write / move are performed on [only for multiple tapes],"},
	    {PState + "6",	"String",	"5. Symbol to write to tape at this position,"},
	    {PState + "7",	"String",	"6. Direction to move tape head ([L]eft, [R]ight or [N]one),"},
	    {PState + "8",	"String",	"7. New state name. (will halt if the specified state isn't defined)"},
	    {PBreakPoint,	"String",	"Comma delimited list of states to break on."},
	    {PAutoStart,	"anything",	"Automatically start program [default: don't]."},
	    {PBGColour,		"String",	"Background colour, in HTML-style rgb form."},
	    {PButtonBG,		"String",	"Button foreground colour, in HTML-style rgb form."},
	    {PDefSpeed,		"int",		"Default speed [0 = step, 4 = crank, default: 1]."},
	    {PDefSpeed, 	"anything",	"Debug (print extra information on stderr)."},
	    {PThreeD,		"anything",	"Render user-interface in 3D."},
	    {PDebug,		"anything",	"Dump error messages to the console."}
	};

	return info;
    }
   
    public void init() {
	initTM();
	setupGUI();
    }

    // Read parameters
    private void initTM() {
	String s;
	int i, j, k;

	// Set background colour
	bg = ((s = getParameter(PBGColour)) != null) ? new Color(Integer.parseInt(s, 16)) : Color.lightGray;
	setBackground(bg);

	// Button background colour
	buttonBG = ((s = getParameter(PButtonBG)) != null) ? new Color(Integer.parseInt(s, 16)) : Color.lightGray;

	// AutoStart?
	autoStart = (getParameter(PAutoStart) != null) ? true : false;

	// Debug?
	debug = (getParameter(PDebug) != null) ? true : false;

	// Render user-interface in 3D? 
	threeD = (getParameter(PThreeD) != null) ? true : false;

	// Get the default speed
	defSpeed = ((s = getParameter(PDefSpeed)) != null) ? Integer.parseInt(s) : 1;

	// Read in the alphabet
	if((s = getParameter(PAlphabet)) != null && s.length() > 0)
	    alphabet = (s.charAt(0) == '\\') ? s.substring(1, s.length()) : s;

	// Create the tape head(s)
	k = ((s = getParameter(PNumTapes)) != null) ? Integer.parseInt(s) : 1;
	tapeHeads = new TapeHead[k];
	for(int t = 1; t <= k; t++) {
	    i = ((s = getParameter(PTapeOffset + t)) != null) ? Integer.parseInt(s) : 0;
	    j = ((s = getParameter(PHeadOffset + t)) != null) ? Integer.parseInt(s) : 0;
	    if((s = getParameter(PTapeString + t)) == null)
		dPrint("TMApplet: no init for tape #" + t);
	    tapeHeads[t - 1] = new TapeHead(this, s, i, j);
	}

	// Read in state transition table
	states = new States(this);
	for(i = 0; true; i++) {
	    s = "State" + i;
	    String st;

	    if((st = getParameter(s)) != null)
		states.newState(st);
	    else
		break;
	}

	// Parse the break point list
	if((s = getParameter(PBreakPoint)) != null)
	    states.parseBP(s);
    }

    private void setupGUI() {
	GridBagLayout gridbag = new GridBagLayout();
	GridBagConstraints c = new GridBagConstraints();

	setLayout(gridbag);

	// Put a few pixels between each component
	c.insets = new Insets(2, 2, 2, 2);
	c.fill = GridBagConstraints.BOTH;

	// states
	c.gridx = 0;
	c.gridy = 0;
	c.gridwidth = 1;
	c.gridheight = 1;
	c.weightx = 1;
	c.weighty = 1;
	gridbag.setConstraints(states, c);
	add(states);

	// control panel
	controlPanel = new ControlPanel(this, defSpeed, buttonBG);
	c.gridx = 1;
	c.gridy = 0;
	c.gridwidth = 1;
	c.gridheight = 1;
	c.weighty = 0;
	gridbag.setConstraints(controlPanel, c);
	add(controlPanel);

	// tape heads
	c.gridx = 0;
	c.gridy = 1;
	c.gridwidth = 2;
	c.gridheight = 1;
	c.weighty = 0;
	for(int i = 0; i < tapeHeads.length; c.gridy++, i++) {
	    gridbag.setConstraints(tapeHeads[i], c);
	    add(tapeHeads[i]);
	}

	validate();
    }

    public void start() {
    	if(autoStart) {
    	    autoStart = false;
    	    if(thread == null) {
    		thread = new Thread(this);
		thread.start();
	    }
	}
    }

    public void stop() {
	thread = null;
    }

    public void run() {
	controlPanel.activeButtons(true, states.Halted());
	// Low priority = responsive GUI
	Thread.currentThread().setPriority(Thread.MIN_PRIORITY);

	if(controlPanel.speed() == 0)
	    states.transition();
	else
	    while(states.transition() && thread != null && controlPanel.speed() != 0 && !states.chkBP()) {
		Thread.yield();
	    }

	// Update drawn tape(s) after a crank
	if(controlPanel.speed() == 4)
	    for(int i = 0; i < tapeHeads.length; i++)
		tapeHeads[i].refresh();

	thread = null;
	controlPanel.activeButtons(false, states.Halted());
    }

    public void paint(Graphics g) {
	update(g);
    }

    public void update(Graphics g) {
	raised3DBorder(g, getSize());
    }

// General operations
    public void lowered3DBorder(Graphics g, Dimension d) {
	g.setColor(Color.black);
        if(threeD) {
	    g.drawLine(0, 0, d.width - 1, 0);
	    g.drawLine(0, 0, 0, d.height - 1);
	    g.setColor(Color.white);
	    g.drawLine(d.width - 1, 0, d.width - 1, d.height - 1);
	    g.drawLine(0, d.height - 1, d.width - 1, d.height - 1);
    //	    g.draw3DRect(0, 0, d.width - 1, d.height - 1, false);
        } else
            g.drawRect(0, 0, d.width - 1, d.height - 1);
    }

    public void flat3DBorder(Graphics g, Dimension d) {
	g.setColor(Color.black);
        g.drawRect(0, 0, d.width - 1, d.height - 1);
    }

    public void raised3DBorder(Graphics g, Dimension d) {
	g.setColor(Color.black);
        if(threeD) {
    	    g.drawLine(d.width - 1, 0, d.width - 1, d.height - 1);
	    g.drawLine(0, d.height - 1, d.width - 1, d.height - 1);
	    g.setColor(Color.white);
	    g.drawLine(0, 0, d.width - 1, 0);
	    g.drawLine(0, 0, 0, d.height - 1);
    //	    g.draw3DRect(0, 0, d.width - 1, d.height - 1, true);
        } else
            g.drawRect(0, 0, d.width - 1, d.height - 1);
    }

    // Alphabet operations
    public char int2sym(int i) {
	char ch = alphabet.charAt(i);
	if(ch < 0) {
	    dPrint("Symbol #" + i + " not found in alphabet.");
	}
	return ch;
    }

    public int sym2int(char ch) {
	int i;

	if((i = alphabet.indexOf(ch)) == -1)
	    dPrint("Character '" + ch + "' not found in alphabet.");
	return i;
    }

    public int alphabetLength() {
	return alphabet.length();
    }

    /*
     * Debugging (of Turing Machines) routines.
     */
    void dPrint(String s) {
	if(debug) {
	    System.err.println(s);
	}
    }
}
