#
# TMParse: parse a spec for a Turing machine.
#
# Peter Gammie - January, 1999 - peteg@unsw.edu.au
# (C) 1999 purely for the purposes of GPL'ing it.
#
package TMParse;

use FileHandle;
use strict;

# Constants
my $halt = 1;
my ($moveNONE, $moveLEFT, $moveRIGHT) = (0 .. 2);

############
# OO rubbish

# Constructor
#  new(FileHandle)
sub new {
  my $proto = shift(@_);
  my $class = ref($proto) || $proto;
  my $self = {};
  (($#_ == 0) && ($self->{fh} = shift(@_))) or die "Constructor called without a filehandle.";

  $self->{STATES} = [];		# State table (3D)
  $self->{STATENUMBERS} = {};	# Map names -> numbers
  $self->{SYMBOLNAMES} = [];	# Map numbers -> names
  $self->{STATEDEFINED} = "";	# Bit = 1 if state number is defined
  $self->{NUMSTATES} = 0;
  $self->{METADATA} = {
		       "alphabet"	=> " 0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ",
		       "appletBG"	=> "",
		       "archive"	=> "../TMApplet.jar",
		       "autoStart"	=> "",
		       "code"	       	=> "TMApplet.TMApplet",
		       "codebase"	=> "",
		       "debug"		=> "",
		       "heading"	=> "",
		       "headOffset"	=> [],
		       "height"		=> "140",
		       "htmlBG"		=> "#FFFFFF",
		       "htmlFG"		=> "",
		       "numTapes"	=> "1",
		       "preHTML"	=> "",
		       "postHTML"	=> "",
		       "tapeOffset"	=> [],
		       "tapeString"	=> [],
		       "numTapes"	=> "1",
		       "threeD"		=> "true",
		       "title" 		=> "",
		       "width" 		=> "600"
		      };
  bless($self, $class);
  return($self);
}

# Auxiliary func
sub _toUNARY {
  my $number = shift(@_);
  my $output = "";

  for(; $number > 0; $number--) {
    $output .= "0";
  }
  return $output;
}


# ADT-ise the state symbol table
sub name2num {
  my $self = shift(@_);
  my $name = shift(@_);
  my $define = shift(@_);
  my $num;

  if(!defined($num = $self->{STATENUMBERS}->{$name})) {
#    print "name2num: '$name' first time -> $numStates.\n";
    $self->{SYMBOLNAMES}->[$self->{NUMSTATES}] = $name;
    $self->{STATENUMBERS}->{$name} = $self->{NUMSTATES};
    $num = $self->{NUMSTATES}++;
  }

  if(defined($define)) {
    # Indicate that the state is valid (non-halting)
#    print "state #$num defined\n";
    vec($self->{STATEDEFINED}, $num, 1) = 1;
  }

  return($num);
}

sub num2name {
  my $self = shift(@_);
  my $num = shift(@_);

  return($self->{SYMBOLNAMES}->[$num]);
}

# ...and the symbol one as well
sub sym2num {
  my $self = shift(@_);
  my $symbol = shift(@_);

  return(index($self->{METADATA}->{"alphabet"}, $symbol));
}

sub num2sym {
  my $self = shift(@_);
  my $num = shift(@_);

  return(substr($self->{METADATA}->{"alphabet"}, $num, 1));
}

# Parse file
sub run {
  my $self = shift(@_);
  my $fh = $self->{fh};

 OUTER: while(<$fh>) {
    # Handle special cases first
    if(/%(preHTML|postHTML)/) {
      print "found: $1\n";
      while(<$fh>) {
	if(/%/) {
	  next OUTER;
	} else {
	  $self->{METADATA}->{$1} .= $_;
	}
      }
      next OUTER;
    }

    if(/%(headOffset|tapeOffset|tapeString)(\d*)\s*"(.*)"/) {
      print "found: $1$2\n";
      $self->{METADATA}->{$1}->[($2 ne "") ? $2 : 1] = $3;
      next OUTER;
    }

    # Use the binary encoding of another machine (just for the
    # universal machine, really).
    if(/%(tapeString)(\d*)\s*(.*)/) {
      print "found: $1$2 (parsing $3)\n";
      $self->{GENTAPE} = TMParse->new(FileHandle->new($3, "r"));
      $self->{GENTAPE}->run();
      my ($tape, $alphabetMap) = $self->{GENTAPE}->genTape();
      $self->{METADATA}->{$1}->[($2 ne "") ? $2 : 1] = $tape;
      $self->{METADATA}->{'alphabetMap'} = $alphabetMap;
      next OUTER;
    }

    # Generic options
    if(/%(\w*?)\s*"(.*)"/) {
      print "found: $1\n";
      $self->{METADATA}->{$1} = $2;
      next OUTER;
    }

    # 7-tuples: use different tapes for read and write/move operations
    if(/\((.+?),(\d+?),(.),(\d+?),(.),(.),(.+?)\)/) {
      if($self->{METADATA}->{"numTapes"} == 1) {
	die("Using 7-tuples requires a multi-tape machine.");
      }
      print "7-tuple: ($1,$2,$3,$4,$5,$6,$7)\n";
      $self->{STATES}->[$self->name2num($1, 1)][$2][$self->sym2num($3)] = [$1, $2, $3, $4, $5, $6, $7];
      next OUTER;
    }

    # 6-tuples: use the same tape for all operations
    if(/\((.+?),(\d+?),(.),(.),(.),(.+?)\)/) {
      if($self->{METADATA}->{"numTapes"} == 1) {
	die("Using 6-tuples requires a multi-tape machine.");
      }
      print "6-tuple: ($1,$2,$3,$4,$5,$6) (expanding to a 7-tuple)\n";
      $self->{STATES}->[$self->name2num($1, 1)][$2][$self->sym2num($3)] = [$1, $2, $3, $2, $4, $5, $6];
      next OUTER;
    }

    # 5-tuples require a single-tape machine
    if(/\((.+?),(.),(.),(.),(.+?)\)/) {
      if($self->{METADATA}->{"numTapes"} > 1) {
	die("Using 5-tuples requires a single tape machine.");
      }
      print "5-tuple: ($1,$2,$3,$4,$5)\n";
      $self->{STATES}->[$self->name2num($1, 1)][1][$self->sym2num($2)] = [$1, $2, $3, $4, $5];
      next OUTER;
    }
  }

  if($self->{NUMSTATES} == 0) {
    die("Must have at least a few state transition tuples.\n");
  }
}

#######################################################################

# Generate HTML data
sub genHTML {
  my $self = shift(@_);

  # Generate state table for the applet
  for(my ($counter, $stateCount) = (0, 0); $stateCount < $self->{NUMSTATES}; $stateCount++) {
    for(my $tapeCount = 1; $tapeCount <= $self->{METADATA}->{"numTapes"}; $tapeCount++) {
      for(my $symCount = 0; $symCount < length($self->{METADATA}->{"alphabet"}); $symCount++) {
	my $state = $self->{STATES}->[$stateCount][$tapeCount][$symCount];
	if(defined($state)) {
	  $self->{METADATA}->{'statetab'} .= "<PARAM name=\"State" . ($counter) . "\" value=\"" . join(",", @{$state}) . "\">\n";
	  $counter++;
	}
      }
    }
  }

  # Generate an HTML table for single tape machines (tricky for multitape)
  if($self->{METADATA}->{'numTapes'} == 1) {
    $self->{METADATA}->{'stateHtab'} = "<TR>\n<TD></TD>\n";
    for(my $symCount = 0; $symCount < length($self->{METADATA}->{'alphabet'}); $symCount++) {
      $self->{METADATA}->{'stateHtab'} .= "<TH ALIGN=\"CENTER\">'" . $self->num2sym($symCount)  . "'</TH>\n";
    }

    for(my $stateCount = 0; $stateCount < $self->{NUMSTATES}; $stateCount++) {
      next if(!vec($self->{STATEDEFINED}, $stateCount, 1));
      $self->{METADATA}->{'stateHtab'} .= "<TR>\n<TH>" . $self->num2name($stateCount) . "</TH>\n";
      for(my $symCount = 0; $symCount < length($self->{METADATA}->{'alphabet'}); $symCount++) {
	my $state = $self->{STATES}->[$stateCount][1][$symCount];
	if(defined($state)) {
	  $self->{METADATA}->{'stateHtab'} .= "<TD ALIGN=\"CENTER\">'$state->[2]',$state->[3],$state->[4]</TD>\n";
	} else {
	  $self->{METADATA}->{'stateHtab'} .= "<TD></TD>\n";
	}
      }
      $self->{METADATA}->{"stateHtab"} .= "</TR>\n";
    }
  } else {
    # For multitape, try to use the simulated machine. Really just for
    # the universal machine.
    if(defined($self->{GENTAPE})) {
      $self->{GENTAPE}->genHTML();
      $self->{METADATA}->{'stateHtab'} = $self->{GENTAPE}->{METADATA}->{'stateHtab'};
    }
  }

  # Other stuff
  for(my $counter = 1; $counter <= $self->{METADATA}->{"numTapes"}; $counter++) {
    if(defined($self->{METADATA}->{"tapeString"}->[$counter])) {
      $self->{METADATA}->{"tapestrings"} .= "<PARAM name=\"TapeString" . $counter . "\" value=\"" . $self->{METADATA}->{"tapeString"}->[$counter] . "\">\n";
    }
    if(defined($self->{METADATA}->{"headOffset"}->[$counter])) {
      $self->{METADATA}->{"headoffsets"} .= "<PARAM name=\"HeadOffset" . $counter . "\" value=\"" . $self->{METADATA}->{"headOffset"}->[$counter] . "\">\n";
    }
    if(defined($self->{METADATA}->{"tapeOffset"}->[$counter])) {
      $self->{METADATA}->{"tapeoffsets"} .= "<PARAM name=\"TapeOffset" . $counter . "\" value=\"" . $self->{METADATA}->{"tapeOffset"}->[$counter] . "\">\n";
    }
  }
}

#######################################################################

# Generate a tape that can be used as input to the universal machine
sub genTape {
  my $self = shift(@_);
  my $output = "";

  # Only works for single tape machines
  if($self->{METADATA}->{"numTapes"} > 1) {
    return($output);
  }

  # Generate state strings
  for(my $stateCount = 0; $stateCount < $self->{NUMSTATES}; $stateCount++) {
    if(vec($self->{STATEDEFINED}, $stateCount, 1)) {
#      print $self->num2name($stateCount) . " defined(" . vec($self->{STATEDEFINED}, $stateCount, 1) . ")\n";
      for(my $symCount = 0; $symCount < length($self->{METADATA}->{'alphabet'}); $symCount++) {
#	print "blah : $stateCount (". num2name($stateCount) . ") ; $numStates\n";

	# If this state tuple doesn't exist, insert a halt transition
	my $state = $self->{STATES}->[$stateCount][1][$symCount];
	if(defined($state)) {
#	  print $self->num2name($stateCount) . " '" . $self->num2sym($symCount) . "' (real)\n";
	  $output .= $self->_printState($state);
#	  print printState($states[$stateCount][1][$symCount]) . "\n";
	} else {
	  print $self->num2name($stateCount) . " '" . $self->num2sym($symCount) . "' (dummy halt)\n";
	  $output .= _printHalt($stateCount, $symCount);
	}
      }
    }
  }

  $output .= "11";

  # Generate tape string
  my $tapeString = $self->{METADATA}->{'tapeString'}->[1];
  if(defined($tapeString)) {
    for(my $tapeIndex = 0; $tapeIndex < length($tapeString); $tapeIndex++) {
      print substr($tapeString, $tapeIndex, 1) . "\n";
      $output .= _toUNARY($self->sym2num(substr($tapeString, $tapeIndex, 1))) . "1";
    }
  } else {
    $output .= "1";
  }

  # Append the head offset
#  print FTAPE toUNARY($vars{"headOffset"}->[0]) . "1";

  # Generate the alphabet mapping, for the universal machine
  my $alphabetMap = "";
  for(my $i = 0; $i < length($self->{METADATA}->{"alphabet"}); $i++) {
    $alphabetMap .= "<TR>\n<TD>'" . substr($self->{METADATA}->{"alphabet"}, $i, 1) . "'</TD><TD>'" . _toUNARY($i) . "'</TD>\n</TR>\n";
  }

  return($output, $alphabetMap);
}

sub _printState {
  my $self = shift(@_);
  my @state = @{shift(@_)};
  my $output;
  my $move;

  # state numbers start at 0
  $output = "1" . _toUNARY($self->name2num($state[0]) + $halt + 1);
  $output .= "1" . _toUNARY($self->sym2num($state[1]));
  $output .= "1" . _toUNARY($self->sym2num($state[2]));

  if($state[3] =~ /L/i) {
    $move = $moveLEFT;
  } else {
    if($state[3] =~ /R/i) {
      $move = $moveRIGHT;
    } else {
      $move = $moveNONE;
    }
  }

  $output .= "1" . _toUNARY($move);

  # Ensure destination state is defined
  if(vec($self->{STATEDEFINED}, $self->name2num($state[4]), 1)) {
    $output .= "1" . _toUNARY($self->name2num($state[4]) + $halt + 1);
  } else {
    $output .= "1" . _toUNARY($halt);
  }
  return($output);
}

sub _printHalt {
  my $state = shift(@_);
  my $symbolIndex = shift(@_);
  my $output;

  $output = "1" . _toUNARY($state + $halt);
  $output .= "1" . _toUNARY($symbolIndex);
  $output .= "1" . _toUNARY($symbolIndex);
  $output .= "1" . _toUNARY($moveNONE);
  $output .= "1" . _toUNARY($halt);
  return($output);
}

1;  # so the require or use succeeds
